
local M = {}

if sys.get_config("script.shared_state") ~= "1" then
	error("rendercam - 'shared_state' setting in game.project (under 'Script') must be enabled for rendercam to work.", 0)
end

-- Localized stuff: (for a small speed boost)
local sqrt = math.sqrt
local min = math.min
local atan = math.atan
local floor = math.floor
local deg, rad = math.deg, math.rad
local matrix4_orthographic = vmath.matrix4_orthographic
local matrix4_perspective = vmath.matrix4_perspective
local m_inv, ortho_inv = vmath.inv, vmath.ortho_inv
local get_world_transform = go.get_world_transform

local function round(x)
	return floor(x + 0.5)
end

-- Constants:
local EXPAND_VIEW = hash("expandView")
local FIXED_AREA = hash("fixedArea")
local FIXED_WIDTH = hash("fixedWidth")
local FIXED_HEIGHT = hash("fixedHeight")
local CONTEXT_KEY = 3700146495

-- Public Variables:
M.configW = sys.get_config("display.width", 960) -- Original window width/height from game.project.
M.configH = sys.get_config("display.height", 480)
M.winW, M.winH = M.configW, M.configH
M.current = nil -- The current camera.
M.debug = {
	name = sys.get_config("rendercam.debug_camera_name", "0"),
	stats = sys.get_config("rendercam.debug_camera_stats", "0"),
	position = sys.get_config("rendercam.debug_camera_position", "0"),
}
for k,v in pairs(M.debug) do
	M.debug[k] = v == "1" and true or false -- Convert to boolean.
end

local cameras = {}

function M.get_zoom_for_resize(scaleMode, newW, newH, oldW, oldH)
	if scaleMode == EXPAND_VIEW then
		return 1
	elseif scaleMode == FIXED_AREA then
		local newArea = newW * newH
		local oldArea = oldW * oldH
		return sqrt(newArea / oldArea) -- Zoom is the scale on both axes, hence the square root.
	elseif scaleMode == FIXED_WIDTH then
		return newW / oldW
	elseif scaleMode == FIXED_HEIGHT then
		return newH / oldH
	end
end

-- Update the camera's viewport to fit within the specified rect.
function M.update_cam_viewport(self, x, y, w, h)
	local vp = self.viewport
	if self.fixedAspectRatio then
		local vpw = min(w, h * self.aspectRatio)
		local vph = vpw / self.aspectRatio
		local extraX, extraY = w - vpw, h - vph
		-- Keep everything as integers to prevent subpixel jitter when animating aspect ratio.
		local ox = round(extraX * self.viewportAlign.x) -- Letterbox offset within rect.
		local oy = round(extraY * self.viewportAlign.y)
		local _ox = round(extraX * (1 - self.viewportAlign.x)) -- Letterbox offset on the opposite corner.
		local _oy = round(extraY * (1 - self.viewportAlign.y))
		vp.x, vp.y = x + ox, y + oy
		vp.w, vp.h = w - ox - _ox, h - oy - _oy -- Width/height is the remaining space.
	else
		vp.x, vp.y, vp.w, vp.h = x, y, w, h
	end
end

function M.get_fov(viewDist, a)
	return deg(atan(a / viewDist) * 2) -- opp / adj
end

function M.get_transform(self)
	return get_world_transform(self.offsetURL)
end

function M.get_view(self, transform)
	local transform = transform or self.transform or get_world_transform(self.offsetURL)
	return ortho_inv(transform)
end

function M.get_projection(self)
	if self.orthographic then
		local hw, hh = self.viewArea.x/2, self.viewArea.y/2
		return matrix4_orthographic(-hw, hw, -hh, hh, self.nearZ, self.farZ)
	else
		local aspectRatio = self.viewArea.x / self.viewArea.y
		return matrix4_perspective(rad(self.fov), aspectRatio, self.nearZ, self.farZ)
	end
end

function M.update_cam_window(self, x, y, newW, newH, oldW, oldH)
	local vp = self.viewport
	local oldVpW, oldVpH = vp.w, vp.h
	M.update_cam_viewport(self, x, y, newW, newH)
	local z = M.get_zoom_for_resize(self.scaleMode, vp.w, vp.h, oldVpW, oldVpH) -- Zoom based on viewport size change.
	self.viewScale = self.viewScale * z -- It's relative zoom.
	self.viewArea.x = self.viewport.w / (self.viewScale * self.zoom) -- New viewArea == zoomed viewport area.
	self.viewArea.y = self.viewport.h / (self.viewScale * self.zoom)
	if not self.orthographic then
		-- Must use Y, vmath.matrix4_perspective uses vertical FOV.
		self.fov = M.get_fov(self.viewDistance, self.viewArea.y * 0.5) -- View triangle is with half-Y - from center to edge.
	end
	self.projection = M.get_projection(self) -- Window resize happens after update, so this is necessary.
end

function M.window_resized(newW, newH, oldW, oldH) -- From render script. NOT called on engine start.
	for i,cam in ipairs(cameras) do
		cam:update_window(0, 0, newW, newH, oldW, oldH)
	end
	M.winW, M.winH = newW, newH
end

 -- Update the view matrices of all cameras. Should call in render script update.
function M.update_camera_transforms()
	local oldContext = _G[CONTEXT_KEY] -- Questionable practice? Yes. Solves all problems easily? Yes.
	for i,cam in ipairs(cameras) do
		if cam.enabled or cam.updateWhenDisabled then
			_G[CONTEXT_KEY] = cam
			cam.transform = M.get_transform(cam)
			cam.view = M.get_view(cam)
		end
		cam._toScreen = nil
		cam._toWorld = nil
	end
	_G[CONTEXT_KEY] = oldContext
end

function M.camera_apply(self) -- Can only be called from the render script.
	local vp = self.viewport
	render.set_viewport(vp.x, vp.y, vp.w, vp.h)
	render.set_view(self.view)
	render.set_projection(self.projection)
end

function M.camera_enable(self)
	if M.current then  M.camera_disable(M.current)  end
	self.enabled = true
	M.current = self
end

function M.get_camera(url)
	for i,cam in ipairs(cameras) do
		if cam.url == url then
			return cam
		end
	end
end

local function register_camera(self)
	table.insert(cameras, self)
end

local function unregister_camera(self)
	for i,cam in ipairs(cameras) do
		if cam == self then
			table.remove(cameras, i)
			break
		end
	end
end

function M.camera_disable(self)
	self.enabled = false
	M.current = nil
end

function M.camera_init(self)
	register_camera(self)
	if not self.useViewArea then
		self.viewArea.x, self.viewArea.y = M.configW, M.configH
	end
	M.update_cam_viewport(self, 0, 0, self.viewArea.x, self.viewArea.y)
	self:update_window(0, 0, M.configW, M.configH, self.viewArea.x, self.viewArea.y)
	self:update_window(0, 0, M.winW, M.winH, self.viewArea.x, self.viewArea.y)
	if self.enabled then  M.camera_enable(self)  end
end

function M.camera_final(self)
	unregister_camera(self)
	if self.enabled then  M.camera_disable(self)  end
end

local v4 = vmath.vector4()

function M.world_to_screen(self, wPos, isDelta)
	self._toScreen = self._toScreen or self.projection * self.view
	local sPos = v4
	sPos.x, sPos.y, sPos.z, sPos.w = wPos.x, wPos.y, wPos.z, 1
	if isDelta then
		local pos = self.transform.c3
		sPos.x, sPos.y, sPos.z = pos.x + sPos.x, pos.y + sPos.y, pos.z + sPos.z
	end
	sPos = self._toScreen * sPos -- Transform from World, through View, into Projection space.
	local x, y, z, w = sPos.x, sPos.y, sPos.z, sPos.w
	x, y, z = x/w, y/w, z/w -- Convert to window -1 to +1.
	if isDelta then
		x, y = x/2, y/2 -- Convert to window 0 to 1.
	else
		x, y = x/2 + 0.5, y/2 + 0.5 -- Convert to window 0 to 1.
	end
	x, y = x * M.winW, y * M.winH -- Convert to window pixels.
	-- Include zero so we can send the result directly to vmath.vector3().
	return x, y, 0 -- NOTE: Results are NOT rounded
end

-- Get the points on the near/far planes that match the specified screen position.
function M.screen_to_world_ray(self, x, y)
	self._toWorld = self._toWorld or m_inv(self.projection * self.view)

	x, y = x / M.winW, y / M.winH -- Convert from window pixels to 0 to 1
	x, y = (x - 0.5) * 2, (y - 0.5) * 2 -- Convert to window -1 to +1

	v4.x, v4.y, v4.z, v4.w = x, y, -1, 1
	local nearP = self._toWorld * v4 -- Transform near-point: (x, y, -1, 1)
	v4.z = 1
	local farP = self._toWorld * v4 -- Transform far-point: (x, y, 1, 1)

	local nx, ny, nz, nw = nearP.x, nearP.y, nearP.z, nearP.w
	nx, ny, nz = nx/nw, ny/nw, nz/nw
	local fx, fy, fz, fw = farP.x, farP.y, farP.z, farP.w
	fx, fy, fz = fx/fw, fy/fw, fz/fw
	return nx, ny, nz, fx, fy, fz
end

-- Get the point at the camera's viewDistance.
function M.screen_to_world_2d(self, x, y)
	local nx, ny, nz, fx, fy, fz = M.screen_to_world_ray(self, x, y)
	-- Get viewDistance as a froction between nearZ and farZ
	local t = ( self.viewDistance - self.nearZ) / (self.farZ - self.nearZ)
	-- Lerp each dimension from near-point to far-point.
	local x = nx + (fx - nx) * t
	local y = ny + (fy - ny) * t
	local z = nz + (fz - nz) * t
	return x, y, z
end

return M
